import os
import sys
from setuptools import Extension
from setuptools import setup

import numpy as np
from Cython.Build import cythonize

# skranger project directory
top = os.path.dirname(os.path.abspath(__file__))

# include skranger, ranger, and numpy headers
# requires running buildpre.py to find src in this location
include_dirs = [
    top,
    os.path.join(top, "skranger"),
    os.path.join(top, "skranger", "ranger", "src"),
    os.path.join(top, "skranger", "ranger", "src", "Forest"),
    os.path.join(top, "skranger", "ranger", "src", "Tree"),
    os.path.join(top, "skranger", "ranger", "src", "utility"),
    np.get_include(),
]


def find_pyx_files(directory, files=None):
    """Recursively find all Cython extension files.

    :param str directory: The directory in which to recursively crawl for .pyx files.
    :param list files: A list of files in which to append discovered .pyx files.
    """
    if files is None:
        files = []
    for file in os.listdir(directory):
        path = os.path.join(directory, file)
        if os.path.isfile(path) and path.endswith(".pyx"):
            files.append(path.replace(os.path.sep, ".")[:-4])
        elif os.path.isdir(path):
            find_pyx_files(path, files)
    return files


def create_extension(module_name):
    """Create a setuptools build extension for a Cython extension file.

    :param str module_name: The name of the module
    """
    path = module_name.replace(".", os.path.sep) + ".pyx"
    return Extension(
        module_name,
        sources=[path],
        include_dirs=include_dirs,
        language="c++",
        extra_compile_args=["-std=c++11", "-Wall"],
        extra_link_args=["-std=c++11", "-g"],
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    )


ext_modules = [create_extension(name) for name in find_pyx_files("skranger")]


def build_ext():
    setup(
        name="skranger",
        ext_modules=cythonize(
            ext_modules,
            gdb_debug=False,
            force=True,
            annotate=False,
            compiler_directives={"language_level": "3"},
        ),
    )


if "clean" in sys.argv or "build_ext" in sys.argv:
    build_ext()


def build(setup_kwargs):
    build_ext()



"""Cython definition file for C++ classes from ranger, skranger, and C++ std."""
from libcpp cimport bool
from libcpp.memory cimport unique_ptr
from libcpp.string cimport string
from libcpp.vector cimport vector


# Wrap stringstream for when verbose == False
cdef extern from "<sstream>" namespace "std":
    cdef cppclass stringstream:
        stringstream() except +

# Wrap cout/iostream for when verbose == True
# https://stackoverflow.com/questions/30984078/cython-working-with-c-streams
cdef extern from "<iostream>" namespace "std":
    ostream cout
    cdef cppclass ostream:
        ostream& write(const char*, int) except +

# Enums required as input for the ForestClassifier
cdef extern from "./ranger/src/globals.h" namespace "ranger":
    cpdef enum MemoryMode:
        MEM_DOUBLE = 0,
        MEM_FLOAT = 1,
        MEM_CHAR = 2

cdef extern from "./ranger/src/globals.h" namespace "ranger":
    # the c++ code is mis-ordered also
    cpdef enum ImportanceMode:
        IMP_NONE = 0,
        IMP_GINI = 1,
        IMP_PERM_BREIMAN = 2,
        IMP_PERM_LIAW = 4,
        IMP_PERM_RAW = 3,
        IMP_GINI_CORRECTED = 5,
        IMP_PERM_CASEWISE = 6

cdef extern from "./ranger/src/globals.h" namespace "ranger":
    cpdef enum SplitRule:
        LOGRANK = 1,
        AUC = 2,
        AUC_IGNORE_TIES = 3,
        MAXSTAT = 4,
        EXTRATREES = 5,
        BETA = 6,
        HELLINGER = 7

cdef extern from "./ranger/src/globals.h" namespace "ranger":
    cpdef enum PredictionType:
        RESPONSE = 1,
        TERMINALNODES = 2

cdef extern from "./ranger/src/globals.h" namespace "ranger":
    cpdef enum TreeType:
        TREE_CLASSIFICATION = 1,
        TREE_REGRESSION = 3,
        TREE_SURVIVAL = 5,
        TREE_PROBABILITY = 9

# Inherit from the Data class so that we can pass numpy array based data objects
cdef extern from "./ranger/src/utility/Data.cpp":
    pass

cdef extern from "./ranger/src/utility/Data.h" namespace "ranger":
    cdef cppclass Data:
        Data() except +
        double get_x(size_t row, size_t col)
        double get_y(size_t row, size_t col)
        size_t getVariableID(char* variable_name)
        void reserveMemory(size_t y_cols)
        void set_x(size_t col, size_t row, double value, bool& error)
        void set_y(size_t col, size_t row, double value, bool& error)

        size_t getSnp(size_t row, size_t col, size_t col_permuted)
        size_t getPermutedSampleID(size_t sampleID)
        size_t getUnpermutedVarID(size_t varID)

# Custom Data child class for numpy
cdef extern from "DataNumpy.h" namespace "ranger":
    cdef cppclass DataNumpy(Data):
        DataNumpy() except +
        DataNumpy(
            double* x,
            double* y,
            vector[string] variable_names,
            size_t num_rows,
            size_t num_cols,
            size_t num_cols_y
        )

cdef extern from "./ranger/src/Tree/Tree.cpp":
    pass

cdef extern from "./ranger/src/Tree/Tree.h" namespace "ranger":
    cdef cppclass Tree:
        Tree() except +
        Tree(
            vector[vector[size_t]]& child_nodeIDs,
            vector[size_t]& split_varIDs,
            vector[double]& split_values
        )

cdef extern from "./ranger/src/Tree/TreeClassification.cpp":
    pass

cdef extern from "./ranger/src/Tree/TreeClassification.h" namespace "ranger":
    cdef cppclass TreeClassification(Tree):
        TreeClassification() except +
        TreeClassification(
            vector[double]* class_values,
            vector[unsigned int]* response_classIDs,
            vector[vector[size_t]]* sampleIDs_per_class,
            vector[double]* class_weights
        )

cdef extern from "./ranger/src/Tree/TreeSurvival.cpp":
    pass

cdef extern from "./ranger/src/Tree/TreeSurvival.h" namespace "ranger":
    cdef cppclass TreeSurvival(Tree):
        TreeSurvival() except +
        TreeSurvival(
            vector[double]* unique_timepoints,
            vector[size_t]* response_timepointIDs,
        )

cdef extern from "./ranger/src/Tree/TreeRegression.cpp":
    pass

cdef extern from "./ranger/src/Tree/TreeRegression.h" namespace "ranger":
    cdef cppclass TreeRegression(Tree):
        TreeRegression() except +
        TreeRegression(
            vector[vector[size_t]]& child_nodeIDs,
            vector[size_t]& split_varIDs,
        )

cdef extern from "./ranger/src/Tree/TreeProbability.cpp":
    pass

cdef extern from "./ranger/src/Tree/TreeProbability.h" namespace "ranger":
    cdef cppclass TreeProbability(Tree):
        TreeProbability() except +
        TreeProbability(
            vector[double]* class_values,
            vector[unsigned int]* response_classIDs,
            vector[vector[size_t]]* sampleIDs_per_class,
            vector[double]* class_weights
        )

cdef extern from "./ranger/src/Forest/Forest.cpp":
    pass

cdef extern from "./ranger/src/Forest/Forest.h" namespace "ranger":
    cdef cppclass Forest:
        Forest() except +
        void init(
            MemoryMode memory_mode,
            unique_ptr[DataNumpy] input_data,
            int mtry,
            char* output_prefix,
            int num_trees,
            int seed,
            int num_threads,
            int importance_mode,
            int min_node_size,
            bool prediction_mode,
            bool sample_with_replacement,
            const vector[string]& unordered_variable_names,
            bool memory_saving_splitting,
            SplitRule splitrule,
            bool predict_all,
            vector[double]& sample_fraction,
            double alpha,
            double minprop,
            bool holdout,
            PredictionType prediction_type,
            int num_random_splits,
            bool order_snps,
            int max_depth,
            const vector[double]& regularization_factor,
            bool regularization_usedepth
        )
        void initR(
            unique_ptr[DataNumpy] input_data,
            int mtry,
            int num_trees,
            ostream* verbose_out,
            int seed,
            int num_threads,
            int importance_mode,
            int min_node_size,
            vector[vector[double]]& split_select_weights,
            const vector[string]& always_split_variable_names,
            bool prediction_mode,
            bool sample_with_replacement,
            const vector[string]& unordered_variable_names,
            bool memory_saving_splitting,
            SplitRule splitrule,
            vector[double]& case_weights,
            vector[vector[size_t]]& manual_inbag,
            bool predict_all,
            bool keep_inbag,
            vector[double]& sample_fraction,
            double alpha,
            double minprop,
            bool holdout,
            PredictionType prediction_type,
            int num_random_splits,
            bool order_snps,
            int max_depth,
            const vector[double]& regularization_factor,
            bool regularization_usedepth,
        ) except +
        void run(bool verbose, bool compute_oob_error)
        void saveToFile()
        vector[vector[vector[size_t]]] getChildNodeIDs()
        const vector[bool]& getIsOrderedVariable()
        unsigned int getMinNodeSize()
        unsigned int getMtry()
        size_t getNumIndependentVariables()
        size_t getNumTrees()
        double getOverallPredictionError()
        const vector[vector[vector[double]]]& getPredictions()
        vector[vector[double]] getSplitValues()
        vector[vector[size_t]] getSplitVarIDs()
        vector[vector[size_t]] getInbagCounts()
        const vector[double]& getVariableImportance()
        const vector[double]& getVariableImportanceCasewise()

cdef extern from "./ranger/src/utility/utility.cpp":
    pass

cdef extern from "./ranger/src/utility/utility.h" namespace "ranger":
    void equalSplit(
        vector[unsigned int]& result,
        unsigned int start,
        unsigned int end,
        unsigned int num_parts
    )

cdef extern from "./ranger/src/Forest/ForestClassification.cpp":
    pass

cdef extern from "./ranger/src/Forest/ForestClassification.h" namespace "ranger":
    cdef cppclass ForestClassification(Forest):
        ForestClassification() except +
        const vector[double]& getClassValues()
        void loadForest(
            size_t num_trees,
            vector[vector[vector[size_t]]]& forest_child_nodeIDs,
            vector[vector[size_t]]& forest_split_varIDs,
            vector[vector[double]]& forest_split_values,
            vector[double]& class_Values,
            vector[bool]& is_ordered_variable,
        )
        void setClassWeights(vector[double]& class_weights)

cdef extern from "./ranger/src/Forest/ForestRegression.cpp":
    pass

cdef extern from "./ranger/src/Forest/ForestRegression.h" namespace "ranger":
    cdef cppclass ForestRegression(Forest):
        ForestRegression() except +
        void loadForest(
            size_t num_trees,
            vector[vector[vector[size_t]]]& forest_child_nodeIDs,
            vector[vector[size_t]]& forest_split_varIDs,
            vector[vector[double]]& forest_split_values,
            vector[bool]& is_ordered_variable,
        )

cdef extern from "./ranger/src/Forest/ForestProbability.cpp":
    pass

cdef extern from "./ranger/src/Forest/ForestProbability.h" namespace "ranger":
    cdef cppclass ForestProbability(Forest):
        ForestProbability() except +
        vector[double]& getClassValues()
        vector[vector[vector[double]]] getTerminalClassCounts()
        void setClassWeights(vector[double]& class_weights)
        void loadForest(
            size_t num_trees,
            vector[vector[vector[size_t]]]& forest_child_nodeIDs,
            vector[vector[size_t]]& forest_split_varIDs,
            vector[vector[double]]& forest_split_values,
            vector[double]& class_Values,
            vector[vector[vector[double]]]& forest_terminal_class_counts,
            vector[bool]& is_ordered_variable,
        )

cdef extern from "./ranger/src/Forest/ForestSurvival.cpp":
    pass

cdef extern from "./ranger/src/Forest/ForestSurvival.h" namespace "ranger":
    cdef cppclass ForestSurvival(Forest):
        ForestSurvival() except +
        vector[vector[vector[double]]] getChf()
        vector[double]& getUniqueTimepoints()
        void loadForest(
            size_t num_trees,
            vector[vector[vector[size_t]]]& forest_child_nodeIDs,
            vector[vector[size_t]]& forest_split_varIDs,
            vector[vector[double]]& forest_split_values,
            vector[vector[vector[double]]]& forest_chf,
            vector[double]& unique_timepoints,
            vector[bool]& is_ordered_variable,
        )



"""Scikit-learn wrapper for ranger classification."""
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.base import ClassifierMixin
from sklearn.exceptions import NotFittedError
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import check_array
from sklearn.utils.validation import check_is_fitted

from skranger import ranger
from skranger.ensemble.base import BaseRangerForest
from skranger.tree.classifier import RangerTreeClassifier


class RangerForestClassifier(BaseRangerForest, ClassifierMixin, BaseEstimator):
    r"""Ranger Random Forest Probability/Classification implementation for sci-kit learn.

    Provides a sklearn classifier interface to the Ranger C++ library using Cython.

    :param int n_estimators: The number of tree classifiers to train
    :param bool verbose: Enable ranger's verbose logging
    :param int/callable mtry: The number of features to split on each node. When a
        callable is passed, the function must accept a single parameter which is the
        number of features passed, and return some value between 1 and the number of
        features.
    :param str importance: One of one of ``none``, ``impurity``, ``impurity_corrected``,
        ``permutation``.
    :param int min_node_size: The minimal node size.
    :param int max_depth: The maximal tree depth; 0 means unlimited.
    :param bool replace: Sample with replacement.
    :param float/list sample_fraction: The fraction of observations to sample. The
        default is 1 when sampling with replacement, and 0.632 otherwise. This can be a
        list of class specific values.
    :param bool keep_inbag: If true, save how often observations are in-bag in each
        tree. These will be stored in the ``ranger_forest_`` attribute under the key
        ``"inbag_counts"``.
    :param list inbag: A list of size ``n_estimators``, containing inbag counts for each
        observation. Can be used for stratified sampling.
    :param str split_rule: One of ``gini``, ``extratrees``, ``hellinger``;
        default ``gini``.
    :param int num_random_splits: The number of random splits to consider for the
        ``extratrees`` splitrule.
    :param str respect_categorical_features: One of ``ignore``, ``order``, ``partition``.
        The default is ``partition`` for the ``extratrees`` splitrule, otherwise the
        default is ``ignore``.
    :param bool scale_permutation_importance: For ``permutation`` importance,
        scale permutation importance by standard error as in (Breiman 2001).
    :param bool local_importance: For ``permutation`` importance, calculate and
        return local importance values as (Breiman 2001).
    :param list regularization_factor: A vector of regularization factors for the
        features.
    :param bool regularization_usedepth: Whether to consider depth in regularization.
    :param bool holdout: Hold-out all samples with case weight 0 and use these for
        feature importance and prediction error.
    :param bool oob_error: Whether to calculate out-of-bag prediction error.
    :param int n_jobs: The number of threads. Default is number of CPU cores.
    :param bool save_memory: Save memory at the cost of speed growing trees.
    :param int seed: Random seed value.
    :param bool enable_tree_details: When ``True``, perform additional calculations
        for detailing the underlying decision trees. Must be enabled for ``estimators_``
        and ``get_estimator`` to work. Very slow.

    :ivar ndarray classes\_: The class labels determined from the fit input ``y``.
    :ivar int n_classes\_: The number of unique class labels from the fit input ``y``.
    :ivar int n_features_in\_: The number of features (columns) from the fit input
        ``X``.
    :ivar list feature_names\_: Names for the features of the fit input ``X``.
    :ivar dict ranger_forest\_: The returned result object from calling C++ ranger.
    :ivar int mtry\_: The mtry value as determined if ``mtry`` is callable, otherwise
        it is the same as ``mtry``.
    :ivar float/list sample_fraction\_: The sample fraction determined by input
        validation
    :ivar list regularization_factor\_: The regularization factors determined by input
        validation.
    :ivar int split_rule\_: The split rule integer corresponding to ranger enum
        ``SplitRule``.
    :ivar bool use_regularization_factor\_: Input validation determined bool for using
        regularization factor input parameter.
    :ivar str respect_categorical_features\_: Input validation determined string
        respecting categorical features.
    :ivar int importance_mode\_: The importance mode integer corresponding to ranger
        enum ``ImportanceMode``.
    :ivar list ranger_class_order\_: The class reference ordering derived from ranger.
    :ivar ndarray feature_importances\_: The variable importances from ranger.
    """

    def __init__(
        self,
        n_estimators=100,
        verbose=False,
        mtry=0,
        importance="none",
        min_node_size=0,
        max_depth=0,
        replace=True,
        sample_fraction=None,
        keep_inbag=False,
        inbag=None,
        split_rule="gini",
        num_random_splits=1,
        respect_categorical_features=None,
        scale_permutation_importance=False,
        local_importance=False,
        regularization_factor=None,
        regularization_usedepth=False,
        holdout=False,
        oob_error=False,
        n_jobs=-1,
        save_memory=False,
        seed=42,
        enable_tree_details=False,
    ):
        self.n_estimators = n_estimators
        self.verbose = verbose
        self.mtry = mtry
        self.importance = importance
        self.min_node_size = min_node_size
        self.max_depth = max_depth
        self.replace = replace
        self.sample_fraction = sample_fraction
        self.keep_inbag = keep_inbag
        self.inbag = inbag
        self.split_rule = split_rule
        self.num_random_splits = num_random_splits
        self.respect_categorical_features = respect_categorical_features
        self.scale_permutation_importance = scale_permutation_importance
        self.local_importance = local_importance
        self.regularization_factor = regularization_factor
        self.regularization_usedepth = regularization_usedepth
        self.holdout = holdout
        self.oob_error = oob_error
        self.n_jobs = n_jobs
        self.save_memory = save_memory
        self.seed = seed
        self.enable_tree_details = enable_tree_details

    @property
    def estimators_(self):
        try:
            check_is_fitted(self)
        except NotFittedError:
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute 'estimators_'"
            ) from None
        if not self.enable_tree_details:
            raise ValueError("enable_tree_details must be True prior to training")
        return [
            RangerTreeClassifier.from_forest(self, idx=idx)
            for idx in range(self.n_estimators)
        ]

    def get_estimator(self, idx):
        """Extract a single estimator tree from the forest.
        :param int idx: The index of the tree to extract.
        """
        check_is_fitted(self)
        if not self.enable_tree_details:
            raise ValueError("enable_tree_details must be True prior to training")
        return RangerTreeClassifier.from_forest(self, idx=idx)

    def fit(
        self,
        X,
        y,
        sample_weight=None,
        class_weights=None,
        split_select_weights=None,
        always_split_features=None,
        categorical_features=None,
    ):
        """Fit the ranger random forest using training data.

        :param array2d X: training input features
        :param array1d y: training input target classes
        :param array1d sample_weight: optional weights for input samples
        :param dict class_weights: A dictionary of outcome classes to weights.
        :param list split_select_weights: Vector of weights between 0 and 1 of
            probabilities to select features for splitting. Can be a single vector or a
            vector of vectors with one vector per tree.
        :param list always_split_features:  Features which should always be selected for
            splitting. A list of column index values.
        :param list categorical_features: A list of column index values which should be
            considered categorical, or unordered.
       """
        self.tree_type_ = 9  # tree_type, TREE_PROBABILITY enables predict_proba

        # Check input
        X, y = self._validate_data(X, y)
        check_classification_targets(y)

        # Check the init parameters
        self._validate_parameters(X, y, sample_weight)

        # Map classes to indices
        y = np.copy(y)
        self.classes_, y = np.unique(y, return_inverse=True)
        self.n_classes_ = len(self.classes_)

        if class_weights is None:
            class_weights = {}
        else:
            try:
                class_weights = {
                    idx: class_weights[k] for idx, k in enumerate(self.classes_)
                }
            except KeyError:
                raise ValueError(
                    "class weights must have a weight for each class"
                ) from None

        # Set X info
        self.feature_names_ = [str(c).encode() for c in range(X.shape[1])]
        self._check_n_features(X, reset=True)

        # Check weights
        sample_weight, use_sample_weight = self._check_sample_weight(sample_weight, X)
        (
            always_split_features,
            use_always_split_features,
        ) = self._check_always_split_features(always_split_features)

        (
            categorical_features,
            use_categorical_features,
        ) = self._check_categorical_features(categorical_features)

        (
            split_select_weights,
            use_split_select_weights,
        ) = self._check_split_select_weights(split_select_weights)

        # Fit the forest
        self.ranger_forest_ = ranger.ranger(
            self.tree_type_,
            np.asfortranarray(X.astype("float64")),
            np.asfortranarray(np.atleast_2d(y).astype("float64").transpose()),
            self.feature_names_,  # variable_names
            self.mtry_,
            self.n_estimators,  # num_trees
            self.verbose,
            self.seed,
            self.n_jobs_,  # num_threads
            True,  # write_forest
            self.importance_mode_,
            self.min_node_size,
            split_select_weights,
            use_split_select_weights,
            always_split_features,  # always_split_variable_names
            bool(always_split_features),  # use_always_split_variable_names
            False,  # prediction_mode
            {},  # loaded_forest
            self.replace,  # sample_with_replacement
            False,  # probability
            categorical_features,  # unordered_variable_names
            use_categorical_features,  # use_unordered_variable_names
            self.save_memory,
            self.split_rule_,
            sample_weight,  # case_weights
            use_sample_weight,  # use_case_weights
            class_weights,
            False,  # predict_all
            self.keep_inbag,
            self.sample_fraction_,
            0.5,  # alpha, ignored because maxstat can't be used on classification
            0.1,  # minprop, ignored because maxstat can't be used on classification
            self.holdout,
            1,  # prediction_type
            self.num_random_splits,
            self.oob_error,
            self.max_depth,
            self.inbag or [],
            bool(self.inbag),  # use_inbag
            self.regularization_factor_,
            False,  # use_regularization_factor
            self.regularization_usedepth,
        )
        self.ranger_class_order_ = np.argsort(
            np.array(self.ranger_forest_["forest"]["class_values"]).astype(int)
        )

        if self.enable_tree_details:
            sample_weight = sample_weight if len(sample_weight) > 0 else np.ones(len(X))
            terminal_node_forest = self._get_terminal_node_forest(X)
            terminal_nodes = np.atleast_2d(terminal_node_forest["predictions"]).astype(
                int
            )
            self._set_leaf_samples(terminal_nodes)
            self._set_node_values(y, sample_weight)
            self._set_n_classes()
        return self

    def predict(self, X):
        """Predict classes from X.

        :param array2d X: prediction input features
        """
        probas = self.predict_proba(X)
        return self.classes_.take(np.argmax(probas, axis=1), axis=0)

    def predict_proba(self, X):
        """Predict probabilities for classes from X.

        :param array2d X: prediction input features
        """
        check_is_fitted(self)
        X = check_array(X)
        self._check_n_features(X, reset=False)

        result = ranger.ranger(
            self.tree_type_,
            np.asfortranarray(X.astype("float64")),
            np.asfortranarray([[]]),
            self.feature_names_,  # variable_names
            self.mtry_,
            self.n_estimators,  # num_trees
            self.verbose,
            self.seed,
            self.n_jobs_,  # num_threads
            False,  # write_forest
            self.importance_mode_,
            self.min_node_size,
            [],  # split_select_weights
            False,  # use_split_select_weights
            [],  # always_split_variable_names
            False,  # use_always_split_variable_names
            True,  # prediction_mode
            self.ranger_forest_["forest"],  # loaded_forest
            self.replace,  # sample_with_replacement
            False,  # probability
            [],  # unordered_feature_names
            False,  # use_unordered_features
            self.save_memory,
            self.split_rule_,
            [],  # case_weights
            False,  # use_case_weights
            {},  # class_weights
            False,  # predict_all
            self.keep_inbag,
            [1],  # sample_fraction
            0.5,  # alpha
            0.1,  # minprop
            self.holdout,
            1,  # prediction_type
            self.num_random_splits,
            self.oob_error,
            self.max_depth,
            self.inbag or [],
            bool(self.inbag),  # use_inbag
            self.regularization_factor_,
            self.use_regularization_factor_,
            self.regularization_usedepth,
        )
        predictions = np.atleast_2d(np.array(result["predictions"]))
        return predictions[:, self.ranger_class_order_]

    def predict_log_proba(self, X):
        """Predict log probabilities for classes from X.

        :param array2d X: prediction input features
        """
        proba = self.predict_proba(X)
        return np.log(proba)

    def _more_tags(self):
        return {
            "_xfail_checks": {
                "check_sample_weights_invariance": "zero sample_weight is not equivalent to removing samples",
            }
        }



import numpy as np
from sklearn.tree._tree import csr_matrix


class Tree:
    """The low-level tree interface.

    Tree objects can be accessed using the ``tree_`` attribute on fitted
    decision tree estimators. Instances of ``Tree`` provide methods and
    properties describing the underlying structure and attributes of the
    tree.
    """

    def __init__(self, ranger_forest):
        self.ranger_forest = ranger_forest

    @property
    def node_count(self):
        """The quantity of nodes in the tree."""
        return len(self.children_left)

    @property
    def capacity(self):
        """The capacity of the node array."""
        return len(self.children_left)

    @property
    def children_left(self):
        """Left children nodes."""
        # sklearn uses -1, ranger uses 0
        return np.array(
            [
                -1 if n == 0 else n
                for n in self.ranger_forest["forest"]["child_node_ids"][0][0]
            ]
        )

    @property
    def children_right(self):
        """Right children nodes."""
        # sklearn uses -1, ranger uses 0
        return np.array(
            [
                -1 if n == 0 else n
                for n in self.ranger_forest["forest"]["child_node_ids"][0][1]
            ]
        )

    @property
    def n_outputs(self):
        """The quantity of outputs of the tree."""
        # single output only
        return 1

    @property
    def n_classes(self):
        """The quantity of classes."""
        return np.array([self.ranger_forest["n_classes"]])

    def get_depth(self):
        """Calculate the maximum depth of the tree."""
        left = self.children_left
        right = self.children_right
        root_node = 0
        return self._get_depth(left, right, root_node)

    def _get_depth(self, left, right, idx):
        if left[idx] == -1:
            return 0
        return 1 + max(
            self._get_depth(left, right, left[idx]),
            self._get_depth(left, right, right[idx]),
        )

    def get_n_leaves(self):
        """Calculate the number of leaves of the tree."""
        left = self.children_left
        right = self.children_right
        root_node = 0
        return self._get_n_leaves(left, right, root_node)

    def _get_n_leaves(self, left, right, idx):
        if left[idx] == -1:
            return 1
        return self._get_n_leaves(left, right, left[idx]) + self._get_n_leaves(
            left, right, right[idx]
        )

    def apply(self, X):
        """Calculate the leaf index for each sample.

        :param array2d X: training input features
        """
        return np.apply_along_axis(self._apply, 1, X)

    def _apply(self, x, idx=None):
        if idx is None:
            idx = 0
            return self._apply(x, idx)
        if self.children_left[idx] == -1:
            return idx
        varid = self.feature[idx]
        val = self.threshold[idx]
        if x[varid] <= val:
            idx = self.children_left[idx]
        else:
            idx = self.children_right[idx]
        return self._apply(x, idx)

    def decision_path(self, X):
        """Calculate the decision path through the tree for each sample.

        :param array2d X: training input features
        """
        if hasattr(X, "values"):  # pd.Dataframe
            Xvalues = X.values
        else:
            Xvalues = X
        paths = [self._decision_path(x) for x in Xvalues]
        rows = [np.ones(len(p), dtype=int) * idx for idx, p in enumerate(paths)]
        rows = np.concatenate(rows, axis=0)
        cols = np.concatenate(paths, axis=0)
        data = np.ones(len(rows), dtype=int)
        return csr_matrix((data, (rows, cols)))

    def _decision_path(self, x, idx=None):
        if idx is None:
            idx = 0
            return [idx] + self._decision_path(x, idx)
        if self.children_left[idx] == -1:
            return []
        varid = self.feature[idx]
        val = self.threshold[idx]
        if x[varid] <= val:
            idx = self.children_left[idx]
        else:
            idx = self.children_right[idx]
        return [idx] + self._decision_path(x, idx)

    @property
    def max_depth(self):
        """Max depth of the tree."""
        return self.get_depth()

    @property
    def feature(self):
        """Variables on which nodes are split."""
        # sklearn uses -2, ranger uses 0
        return np.array(
            [
                -2 if n == 0 else v
                for n, v in zip(
                    self.ranger_forest["forest"]["child_node_ids"][0][0],
                    self.ranger_forest["forest"]["split_var_ids"][0],
                )
            ]
        )

    @property
    def threshold(self):
        """Threshold values on which nodes are split."""
        # sklearn uses -2, ranger uses 0
        return np.array(
            [
                -2 if n == 0 else v
                for n, v in zip(
                    self.ranger_forest["forest"]["child_node_ids"][0][0],
                    self.ranger_forest["forest"]["split_values"][0],
                )
            ]
        )

    @property
    def n_node_samples(self):
        """The number of samples reaching each node."""
        n_samples = [
            len(node) if node else 0
            for node in self.ranger_forest["forest"]["leaf_samples"][0]
        ]
        self._get_n_node_samples(self.children_left, self.children_right, 0, n_samples)
        return np.array(n_samples)

    def _get_n_node_samples(self, left, right, idx, n_samples):
        left_n_node_samples = (
            n_samples[idx]
            if left[idx] == -1
            else self._get_n_node_samples(left, right, left[idx], n_samples)
        )
        right_n_node_samples = (
            n_samples[idx]
            if right[idx] == -1
            else self._get_n_node_samples(left, right, right[idx], n_samples)
        )
        n_samples[idx] = left_n_node_samples + right_n_node_samples
        return n_samples[idx]

    @property
    def weighted_n_node_samples(self):
        """The sum of the weights of the samples reaching each node."""
        weighted_n_samples = self.ranger_forest["forest"]["leaf_weights"][0].copy()
        self._get_n_node_samples(
            self.children_left, self.children_right, 0, weighted_n_samples,
        )
        return np.array(weighted_n_samples)

    @property
    def value(self):
        """The constant prediction value of each node."""
        values = self.ranger_forest["forest"]["node_values"][0]
        return np.reshape(values, (len(values), 1, self.n_classes[0]))



import bisect

import numpy as np
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from skranger.base import RangerMixin


class BaseRangerForest(RangerMixin):
    @property
    def feature_importances_(self):
        try:
            check_is_fitted(self)
        except NotFittedError:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute 'feature_importances_'"
            ) from None
        try:
            return self.ranger_forest_["variable_importance"]
        except KeyError:
            raise ValueError(
                "importance must be set to something other than 'none'"
            ) from None

    def get_importance_pvalues(self):
        """Calculate p-values for variable importances.

        Uses the fast method from Janitza et al. (2016).
        """
        check_is_fitted(self)
        if self.importance != "impurity_corrected":
            raise ValueError(
                "p-values can only be calculated with importance parameter set to 'impurity_corrected'"
            )

        vimp = np.array(self.ranger_forest_["variable_importance"])
        m1 = vimp[vimp < 0]
        m2 = vimp[vimp == 0]

        if len(m1) == 0:
            raise ValueError(
                "No negative importance values found, cannot calculate p-values."
            )
        if len(m2) < 1:
            vimp_dist = np.concatenate((m1, -m1))
        else:
            vimp_dist = np.concatenate((m1, -m1, m2))

        vimp_dist.sort()
        result = []
        for i in range(len(vimp)):
            result.append(bisect.bisect_left(vimp_dist, vimp[i]))
        pval = 1 - np.array(result) / len(vimp_dist)
        return pval



from contextlib import contextmanager

from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import RandomForestRegressor
from sklearn.tree._tree import Tree as SKTree

from skranger.ensemble import RangerForestClassifier
from skranger.ensemble import RangerForestRegressor
from skranger.tree._tree import Tree


def _getclass(klass):
    """Get a __getattribute__ patch to override __class__."""

    def __class__override(self, item):
        if item == "__class__":
            return klass
        return object.__getattribute__(self, item)

    return __class__override


@contextmanager
def shap_patch():
    """Trick shap into thinking skranger objects are sklearn objects."""
    tree_orig = Tree.__getattribute__
    Tree.__getattribute__ = _getclass(SKTree)
    reg_orig = RangerForestRegressor.__getattribute__
    RangerForestRegressor.__getattribute__ = _getclass(RandomForestRegressor)
    cls_orig = RangerForestClassifier.__getattribute__
    RangerForestClassifier.__getattribute__ = _getclass(RandomForestClassifier)
    yield
    Tree.__getattribute__ = tree_orig
    RangerForestRegressor.__getattribute__ = reg_orig
    RangerForestClassifier.__getattribute__ = cls_orig



"""Cython implementation for ranger and Data child classes."""
import sys

import cython
import numpy as np
cimport numpy as np
from cython.operator cimport dereference as deref
from libcpp cimport bool
from libcpp.memory cimport unique_ptr
from libcpp.string cimport string
from libcpp.utility cimport move
from libcpp.vector cimport vector

from skranger cimport ranger_


cdef class DataNumpy:
    """Cython wrapper for DataNumpy C++ class in ``DataNumpy.h``.

    This wraps the Data class in C++, which encapsulates training data passed to the
    random forest classes. It allows us to pass numpy arrays as a ranger-compatible
    Data object.
    """
    cdef unique_ptr[ranger_.DataNumpy] c_data

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __cinit__(self,
        np.ndarray[double, ndim=2, mode="fortran"] x not None,
        np.ndarray[double, ndim=2, mode="fortran"] y not None,
        vector[string] variable_names,
    ):
        cdef size_t num_rows = np.PyArray_DIMS(x)[0]  # in lieu of x.shape
        cdef size_t num_cols = np.PyArray_DIMS(x)[1]
        cdef size_t num_cols_y = np.PyArray_DIMS(y)[1]
        self.c_data.reset(
            new ranger_.DataNumpy(
                &x[0, 0],
                &y[0, 0],
                variable_names,
                num_rows,
                num_cols,
                num_cols_y,
            )
        )

    def get_x(self, size_t row, size_t col):
        return deref(self.c_data).get_x(row, col)

    def get_y(self, size_t row, size_t col):
        return deref(self.c_data).get_y(row, col)

    def reserve_memory(self, size_t y_cols):
        return deref(self.c_data).reserveMemory(y_cols)

    def set_x(self, size_t col, size_t row, double value, bool& error):
        return deref(self.c_data).set_x(col, row, value, error)

    def set_y(self, size_t col, size_t row, double value, bool& error):
        return deref(self.c_data).set_y(col, row, value, error)


cpdef dict ranger(
    ranger_.TreeType treetype,
    np.ndarray[double, ndim=2, mode="fortran"] x,
    np.ndarray[double, ndim=2, mode="fortran"] y,
    vector[string]& variable_names,
    unsigned int mtry,
    unsigned int num_trees,
    bool verbose,
    unsigned int seed,
    unsigned int num_threads,
    bool write_forest,
    ranger_.ImportanceMode importance_mode,
    unsigned int min_node_size,
    vector[vector[double]]& split_select_weights,
    bool use_split_select_weights,
    vector[string]& always_split_variable_names,
    bool use_always_split_variable_names,
    bool prediction_mode,
    dict loaded_forest,
    bool sample_with_replacement,
    bool probability,
    vector[string]& unordered_variable_names,
    bool use_unordered_variable_names,
    bool save_memory,
    ranger_.SplitRule splitrule,
    vector[double]& case_weights,
    bool use_case_weights,
    dict class_weights,
    bool predict_all,
    bool keep_inbag,
    vector[double]& sample_fraction,
    double alpha,
    double minprop,
    bool holdout,
    ranger_.PredictionType prediction_type,
    unsigned int num_random_splits,
    bool oob_error,
    unsigned int max_depth,
    vector[vector[size_t]]& inbag,
    bool use_inbag,
    vector[double]& regularization_factor,
    bool use_regularization_factor,
    bool regularization_usedepth,
):
    """Cython function interface to ranger.
    
    Provides an entrypoint into the ranger C++ code, and returns a result object
    with ranger-specific random forest implementation objects for serializing and
    deserializing random forests. The result object is a python dictionary containing
    forest structures, metadata, and predictions (if in prediction mode). The structure
    of results is (depending on forest type):
    
    {
        "predictions": [[]],
        "num_trees": int,
        "num_independent_variables": int,
        "mtry": int,
        "min_node_size": int,
        "prediction_error": float,
        "inbag_counts": -,
        "unique_death_times": -,
        "variable_importance": -,
        "variable_importance_local": -,
        "forest": {
            "num_trees": int,
            "child_node_ids": [[[]]],
            "split_var_ids": [[]],
            "split_values": [[]],
            "is_ordered": [],
            "class_values": [],
            "terminal_class_counts": [[[]]],
            "cumulative_hazard_function": -,
            "unique_death_times": -,
        },
    }
    """
    # print(locals())
    result = {}

    cdef unique_ptr[ranger_.Forest] forest

    cdef ranger_.ostream* verbose_out

    cdef vector[vector[vector[size_t]]] child_node_ids
    cdef vector[vector[size_t]] split_var_ids
    cdef vector[vector[double]] split_values
    cdef vector[bool] is_ordered
    cdef vector[double] class_values
    cdef vector[double] class_values_
    cdef vector[vector[vector[double]]] cumulative_hazard_function
    cdef vector[double] unique_timepoints
    cdef vector[vector[vector[double]]] terminal_class_counts
    cdef vector[vector[vector[double]]] predictions

    try:
        if not use_split_select_weights:
            split_select_weights.clear()
        if not use_always_split_variable_names:
            always_split_variable_names.clear()
        if not use_unordered_variable_names:
            unordered_variable_names.clear()
        if not use_case_weights:
            case_weights.clear()
        if not use_inbag:
            inbag.clear()
        if not use_regularization_factor:
            regularization_factor.clear()

        if verbose:
            verbose_out = <ranger_.ostream*> &ranger_.cout
        else:
            verbose_out = <ranger_.ostream*> new ranger_.stringstream()

        data = DataNumpy(x, y, variable_names)

        if treetype == ranger_.TreeType.TREE_CLASSIFICATION:
            if probability:
                forest.reset(new ranger_.ForestProbability())
            else:
                forest.reset(new ranger_.ForestClassification())
        elif treetype == ranger_.TreeType.TREE_REGRESSION:
            forest.reset(new ranger_.ForestRegression())
        elif treetype == ranger_.TreeType.TREE_SURVIVAL:
            forest.reset(new ranger_.ForestSurvival())
        elif treetype == ranger_.TreeType.TREE_PROBABILITY:
            forest.reset(new ranger_.ForestProbability())

        deref(forest).initR(
            move(data.c_data),
            mtry,
            num_trees,
            verbose_out,
            seed,
            num_threads,
            importance_mode,
            min_node_size,
            split_select_weights,
            always_split_variable_names,
            prediction_mode,
            sample_with_replacement,
            unordered_variable_names,
            save_memory,
            splitrule,
            case_weights,
            inbag,
            predict_all,
            keep_inbag,
            sample_fraction,
            alpha,
            minprop,
            holdout,
            prediction_type,
            num_random_splits,
            False,  # order_snps
            max_depth,
            regularization_factor,
            regularization_usedepth,
        )

        if prediction_mode:
            child_node_ids = loaded_forest["child_node_ids"]
            split_var_ids = loaded_forest["split_var_ids"]
            split_values = loaded_forest["split_values"]
            is_ordered = loaded_forest["is_ordered"]

            if treetype == ranger_.TreeType.TREE_CLASSIFICATION:
                class_values = loaded_forest["class_values"]
                (<ranger_.ForestClassification*> forest.get()).loadForest(num_trees, child_node_ids, split_var_ids, split_values, class_values, is_ordered)
            elif treetype == ranger_.TreeType.TREE_REGRESSION:
                (<ranger_.ForestRegression*> forest.get()).loadForest(num_trees, child_node_ids, split_var_ids, split_values, is_ordered)
            elif treetype == ranger_.TreeType.TREE_SURVIVAL:
                cumulative_hazard_function = loaded_forest["cumulative_hazard_function"]
                unique_timepoints = loaded_forest["unique_death_times"]
                (<ranger_.ForestSurvival*> forest.get()).loadForest(num_trees, child_node_ids, split_var_ids, split_values, cumulative_hazard_function, unique_timepoints, is_ordered)
            elif treetype == ranger_.TreeType.TREE_PROBABILITY:
                class_values = loaded_forest["class_values"]
                terminal_class_counts = loaded_forest["terminal_class_counts"]
                (<ranger_.ForestProbability*> forest.get()).loadForest(num_trees, child_node_ids, split_var_ids, split_values, class_values, terminal_class_counts, is_ordered)
        else:
            if treetype == ranger_.TreeType.TREE_CLASSIFICATION and len(class_weights) > 0:
                class_values_ = (<ranger_.ForestClassification*> forest.get()).getClassValues()
                class_weights_ = []
                for c in class_values_[:class_values_.size()]:
                    class_weights_.append(class_weights[int(c)])
                (<ranger_.ForestClassification*> forest.get()).setClassWeights(class_weights_)
            elif treetype == ranger_.TreeType.TREE_PROBABILITY and len(class_weights) > 0:
                class_values_ = (<ranger_.ForestProbability*> forest.get()).getClassValues()
                class_weights_ = []
                for c in class_values_[:class_values_.size()]:
                    class_weights_.append(class_weights[int(c)])
                (<ranger_.ForestProbability*> forest.get()).setClassWeights(class_weights_)

        deref(forest).run(verbose, oob_error)

        if use_split_select_weights and importance_mode != ranger_.ImportanceMode.IMP_NONE:
            if verbose_out:
                verbose_out.write("Warning: Split select weights used. Variable importance measures are only comparable for variables with equal weights.\n", 1)

        predictions = deref(forest).getPredictions()
        if predictions.size() == 1:
            if predictions[0].size() == 1:
                result["predictions"] = predictions[0][0]
            else:
                result["predictions"] = predictions[0]
        else:
            result["predictions"] = predictions

        result["num_trees"] = deref(forest).getNumTrees()
        result["num_independent_variables"] = deref(forest).getNumIndependentVariables()
        if treetype == ranger_.TreeType.TREE_SURVIVAL:
            result["unique_death_times"] = (<ranger_.ForestSurvival*> forest.get()).getUniqueTimepoints()
        if not prediction_mode:
            result["mtry"] = deref(forest).getMtry()
            result["min_node_size"] = deref(forest).getMinNodeSize()
            if importance_mode != ranger_.ImportanceMode.IMP_NONE:
                result["variable_importance"] = deref(forest).getVariableImportance()
                if importance_mode == ranger_.ImportanceMode.IMP_PERM_CASEWISE:
                    result["variable_importance_local"] = deref(forest).getVariableImportanceCasewise()
            result["prediction_error"] = deref(forest).getOverallPredictionError()

        if keep_inbag:
            result["inbag_counts"] = deref(forest).getInbagCounts()

        if write_forest:
            forest_object = {
                "num_trees": deref(forest).getNumTrees(),
                "child_node_ids": deref(forest).getChildNodeIDs(),
                "split_var_ids": deref(forest).getSplitVarIDs(),
                "split_values": deref(forest).getSplitValues(),
                "is_ordered": deref(forest).getIsOrderedVariable()
            }

            if treetype == ranger_.TreeType.TREE_CLASSIFICATION:
                class_values_ = (<ranger_.ForestClassification*> forest.get()).getClassValues()
                forest_object["class_values"] = []
                for c in class_values_[:class_values_.size()]:
                    forest_object["class_values"].append(c)
            elif treetype == ranger_.TreeType.TREE_PROBABILITY:
                forest_object["class_values"] = (<ranger_.ForestProbability*> forest.get()).getClassValues()
                forest_object["terminal_class_counts"] = (<ranger_.ForestProbability*> forest.get()).getTerminalClassCounts()
            elif treetype == ranger_.TreeType.TREE_SURVIVAL:
                forest_object["cumulative_hazard_function"] = (<ranger_.ForestSurvival*> forest.get()).getChf()
                forest_object["unique_death_times"] = (<ranger_.ForestSurvival*> forest.get()).getUniqueTimepoints()
            result["forest"] = forest_object

        if not verbose:
            del verbose_out

    except Exception as exc:
        raise exc

    return result



"""Scikit-learn wrapper for ranger survival."""
import typing as t

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array
from sklearn.utils.validation import check_is_fitted

from skranger import ranger
from skranger.tree.base import BaseRangerTree

if t.TYPE_CHECKING:  # pragma: no cover
    from skranger.ensemble.survival import RangerForestSurvival


class RangerTreeSurvival(BaseRangerTree, BaseEstimator):
    r"""Ranger Survival implementation for sci-kit survival.

    Provides a sksurv interface to the Ranger C++ library using Cython.

    :param bool verbose: Enable ranger's verbose logging
    :param int/callable mtry: The number of features to split on each node. When a
        callable is passed, the function must accept a single parameter which is the
        number of features passed, and return some value between 1 and the number of
        features.
    :param str importance: One of one of ``none``, ``impurity``, ``impurity_corrected``,
        ``permutation``.
    :param int min_node_size: The minimal node size.
    :param int max_depth: The maximal tree depth; 0 means unlimited.
    :param bool replace: Sample with replacement.
    :param float sample_fraction: The fraction of observations to sample. The default
        is 1 when sampling with replacement, and 0.632 otherwise.
    :param bool keep_inbag: If true, save how often observations are in-bag in each
        tree. These will be stored in the ``ranger_forest_`` attribute under the key
        ``"inbag_counts"``.
    :param list inbag: A list of size ``n_estimators``, containing inbag counts for each
        observation. Can be used for stratified sampling.
    :param str split_rule: One of ``logrank``, ``extratrees``, ``C``, or ``maxstat``,
        default ``logrank``.
    :param int num_random_splits: The number of random splits to consider for the
        ``extratrees`` splitrule.
    :param float alpha: Significance threshold to allow splitting for the ``maxstat``
        split rule.
    :param float minprop: Lower quantile of covariate distribution to be considered for
        splitting for ``maxstat`` split rule.
    :param str respect_categorical_features: One of ``ignore``, ``order``, ``partition``.
        The default is ``partition`` for the ``extratrees`` splitrule, otherwise the
        default is ``ignore``.
    :param bool scale_permutation_importance: For ``permutation`` importance,
        scale permutation importance by standard error as in (Breiman 2001).
    :param bool local_importance: For ``permutation`` importance, calculate and
        return local importance values as (Breiman 2001).
    :param list regularization_factor: A vector of regularization factors for the
        features.
    :param bool regularization_usedepth: Whether to consider depth in regularization.
    :param bool holdout: Hold-out all samples with case weight 0 and use these for
        feature importance and prediction error.
    :param bool oob_error: Whether to calculate out-of-bag prediction error.
    :param int seed: Random seed value.

    :ivar int n_features_in\_: The number of features (columns) from the fit input ``X``.
    :ivar list feature_names\_: Names for the features of the fit input ``X``.
    :ivar dict ranger_forest\_: The returned result object from calling C++ ranger.
    :ivar int mtry\_: The mtry value as determined if ``mtry`` is callable, otherwise
        it is the same as ``mtry``.
    :ivar float sample_fraction\_: The sample fraction determined by input validation.
    :ivar list regularization_factor\_: The regularization factors determined by input
        validation.
    :ivar list unordered_feature_names\_: The unordered feature names determined by
        input validation.
    :ivar int split_rule\_: The split rule integer corresponding to ranger enum
        ``SplitRule``.
    :ivar bool use_regularization_factor\_: Input validation determined bool for using
        regularization factor input parameter.
    :ivar str respect_categorical_features\_: Input validation determined string
        respecting categorical features.
    :ivar int importance_mode\_: The importance mode integer corresponding to ranger
        enum ``ImportanceMode``.
    :ivar ndarray feature_importances\_: The variable importances from ranger.
    """

    def __init__(
        self,
        verbose=False,
        mtry=0,
        importance="none",
        min_node_size=0,
        max_depth=0,
        replace=True,
        sample_fraction=None,
        keep_inbag=False,
        inbag=None,
        split_rule="logrank",
        num_random_splits=1,
        alpha=0.5,
        minprop=0.1,
        respect_categorical_features=None,
        scale_permutation_importance=False,
        local_importance=False,
        regularization_factor=None,
        regularization_usedepth=False,
        holdout=False,
        oob_error=False,
        seed=42,
    ):
        self.verbose = verbose
        self.mtry = mtry
        self.importance = importance
        self.min_node_size = min_node_size
        self.max_depth = max_depth
        self.replace = replace
        self.sample_fraction = sample_fraction
        self.keep_inbag = keep_inbag
        self.inbag = inbag
        self.split_rule = split_rule
        self.num_random_splits = num_random_splits
        self.alpha = alpha
        self.minprop = minprop
        self.respect_categorical_features = respect_categorical_features
        self.scale_permutation_importance = scale_permutation_importance
        self.local_importance = local_importance
        self.regularization_factor = regularization_factor
        self.regularization_usedepth = regularization_usedepth
        self.holdout = holdout
        self.oob_error = oob_error
        self.seed = seed

    @classmethod
    def from_forest(cls, forest: "RangerForestSurvival", idx: int):
        """Extract a tree from a forest.

        :param RangerForestClassifier forest: A trained RangerForestClassifier instance
        :param int idx: The tree index from the forest to extract.
        """
        # Even though we have a tree object, we keep the exact same dictionary structure
        # that exists in the forests, so that we can reuse the Cython entrypoints.
        # We also copy over some instance attributes from the trained forest.

        # params
        instance = cls(
            verbose=forest.verbose,
            mtry=forest.mtry,
            importance=forest.importance,
            min_node_size=forest.min_node_size,
            max_depth=forest.max_depth,
            replace=forest.replace,
            sample_fraction=forest.sample_fraction,
            keep_inbag=forest.keep_inbag,
            inbag=forest.inbag,
            split_rule=forest.split_rule,
            num_random_splits=forest.num_random_splits,
            alpha=forest.alpha,
            minprop=forest.minprop,
            respect_categorical_features=forest.respect_categorical_features,
            scale_permutation_importance=forest.scale_permutation_importance,
            local_importance=forest.local_importance,
            regularization_factor=forest.regularization_factor,
            regularization_usedepth=forest.regularization_usedepth,
            holdout=forest.holdout,
            oob_error=forest.oob_error,
            seed=forest.seed,
        )
        # forest
        ranger_forest = {}
        for k, v in forest.ranger_forest_.items():
            if k == "forest":
                ranger_forest[k] = {}
                for fk, fv in v.items():
                    if isinstance(fv, list) and len(fv) > 0 and isinstance(fv[0], list):
                        ranger_forest[k][fk] = [fv[idx]]
                    else:
                        ranger_forest[k][fk] = fv
            else:
                ranger_forest[k] = v
        ranger_forest["num_trees"] = 1
        instance.ranger_forest_ = ranger_forest
        # vars
        instance.n_features_in_ = forest.n_features_in_
        instance.feature_names_ = forest.feature_names_
        instance.sample_fraction_ = forest.sample_fraction_
        instance.mtry_ = forest.mtry_
        instance.regularization_factor_ = forest.regularization_factor_
        instance.split_rule_ = forest.split_rule_
        instance.use_regularization_factor_ = forest.use_regularization_factor_
        instance.respect_categorical_features_ = forest.respect_categorical_features_
        instance.importance_mode_ = forest.importance_mode_
        instance.tree_type_ = forest.tree_type_
        instance.event_times_ = forest.event_times_
        return instance

    def fit(
        self,
        X,
        y,
        sample_weight=None,
        split_select_weights=None,
        always_split_features=None,
        categorical_features=None,
    ):
        """Fit the ranger random forest using training data.

        :param array2d X: training input features
        :param array2d y: training input targets, rows of (bool, float)
            representing (survival, time)
        :param array1d sample_weight: optional weights for input samples
        :param list split_select_weights: Vector of weights between 0 and 1 of
            probabilities to select features for splitting. Can be a single vector or a
            vector of vectors with one vector per tree.
        :param list always_split_features:  Features which should always be selected for
            splitting. A list of column index values.
        :param list categorical_features: A list of column index values which should be
            considered categorical, or unordered.
        """
        self.tree_type_ = 5  # tree_type, TREE_SURVIVAL

        # Check input
        X = check_array(X)

        # convert 1d array of 2tuples to 2d array
        # ranger expects the time first, and status second
        # since we follow the scikit-survival convention, we fliplr
        yr = np.fliplr(np.array(y.tolist()))

        # Check the init parameters
        self._validate_parameters(X, y, sample_weight)

        # Set X info
        self.feature_names_ = [str(c).encode() for c in range(X.shape[1])]
        self._check_n_features(X, reset=True)

        # Check weights
        sample_weight, use_sample_weight = self._check_sample_weight(sample_weight, X)
        (
            always_split_features,
            use_always_split_features,
        ) = self._check_always_split_features(always_split_features)

        (
            categorical_features,
            use_categorical_features,
        ) = self._check_categorical_features(categorical_features)

        (
            split_select_weights,
            use_split_select_weights,
        ) = self._check_split_select_weights(split_select_weights)

        # Fit the forest
        self.ranger_forest_ = ranger.ranger(
            self.tree_type_,
            np.asfortranarray(X.astype("float64")),
            np.asfortranarray(yr.astype("float64")),
            self.feature_names_,  # variable_names
            self.mtry_,
            1,  # num_trees
            self.verbose,
            self.seed,
            1,  # num_threads
            True,  # write_forest
            self.importance_mode_,
            self.min_node_size,
            split_select_weights,
            use_split_select_weights,
            always_split_features,  # always_split_variable_names
            use_always_split_features,  # use_always_split_variable_names
            False,  # prediction_mode
            {},  # loaded_forest
            self.replace,  # sample_with_replacement
            False,  # probability
            categorical_features,  # unordered_feature_names
            use_categorical_features,  # use_unordered_features
            False,  # save_memory
            self.split_rule_,
            sample_weight,  # case_weights
            use_sample_weight,  # use_case_weights
            {},  # class_weights
            False,  # predict_all
            self.keep_inbag,
            self.sample_fraction_,
            self.alpha,
            self.minprop,
            self.holdout,
            1,  # prediction_type
            self.num_random_splits,
            self.oob_error,
            self.max_depth,
            self.inbag or [],
            bool(self.inbag),  # use_inbag
            self.regularization_factor_,
            False,  # use_regularization_factor
            self.regularization_usedepth,
        )
        self.event_times_ = np.array(
            self.ranger_forest_["forest"]["unique_death_times"]
        )
        # dtype to suppress warning about ragged nested sequences
        self.cumulative_hazard_function_ = np.array(
            self.ranger_forest_["forest"]["cumulative_hazard_function"], dtype=object
        )
        sample_weight = sample_weight if sample_weight != [] else np.ones(len(X))

        terminal_node_forest = self._get_terminal_node_forest(X)
        terminal_nodes = np.atleast_2d(terminal_node_forest["predictions"]).astype(int)
        self._set_leaf_samples(terminal_nodes)
        self._set_node_values(np.array(y.tolist()), sample_weight)
        self._set_n_classes()
        return self

    def _predict(self, X):
        check_is_fitted(self)
        X = check_array(X)
        self._check_n_features(X, reset=False)

        result = ranger.ranger(
            self.tree_type_,
            np.asfortranarray(X.astype("float64")),
            np.asfortranarray([[]]),
            self.feature_names_,  # variable_names
            self.mtry_,
            1,  # num_trees
            self.verbose,
            self.seed,
            1,  # num_threads
            False,  # write_forest
            self.importance_mode_,
            self.min_node_size,
            [],
            False,  # use_split_select_weights
            [],  # always_split_variable_names
            False,  # use_always_split_variable_names
            True,  # prediction_mode
            self.ranger_forest_["forest"],  # loaded_forest
            self.replace,  # sample_with_replacement
            False,  # probability
            [],  # unordered_feature_names
            False,  # use_unordered_features
            False,  # save_memory
            self.split_rule_,
            [],  # case_weights
            False,  # use_case_weights
            {},  # class_weights
            False,  # predict_all
            self.keep_inbag,
            [1],  # sample_fraction
            self.alpha,
            self.minprop,
            self.holdout,
            1,  # prediction_type
            self.num_random_splits,
            self.oob_error,
            self.max_depth,
            self.inbag or [],
            bool(self.inbag),  # use_inbag
            self.regularization_factor_,
            self.use_regularization_factor_,
            self.regularization_usedepth,
        )
        return result

    def predict_cumulative_hazard_function(self, X):
        """Predict cumulative hazard function.

        :param array2d X: prediction input features
        """
        result = self._predict(X)
        return np.atleast_2d(result["predictions"])

    def predict_survival_function(self, X):
        """Predict survival function.

        :param array2d X: prediction input features
        """
        chf = self.predict_cumulative_hazard_function(X)
        return np.exp(-chf)

    def predict(self, X):
        """Predict risk score.

        :param array2d X: prediction input features
        """
        chf = self.predict_cumulative_hazard_function(X)
        return chf.sum(1)

    def _more_tags(self):
        return {
            "requires_y": True,
            "_xfail_checks": {
                "check_sample_weights_invariance": "zero sample_weight is not equivalent to removing samples",
            },
        }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['skranger', 'skranger.ensemble', 'skranger.tree', 'skranger.utils']

package_data = \
{'': ['*'],
 'skranger': ['ranger/*',
              'ranger/cross_compile/*',
              'ranger/src/*',
              'ranger/src/Forest/*',
              'ranger/src/Tree/*',
              'ranger/src/utility/*',
              'ranger/test/*',
              'ranger/test/testthat/*']}

install_requires = \
['scikit-learn>=0.23.0,<1']

setup_kwargs = {
    'name': 'skranger',
    'version': '0.7.0',
    'description': 'Python bindings for C++ ranger random forests',
    'long_description': 'skranger\n========\n\n|actions| |wheels| |rtd| |pypi| |pyversions|\n\n.. |actions| image:: https://github.com/crflynn/skranger/workflows/build/badge.svg\n    :target: https://github.com/crflynn/skranger/actions\n\n.. |wheels| image:: https://github.com/crflynn/skranger-wheels/workflows/wheels/badge.svg\n    :target: https://github.com/crflynn/skranger-wheels/actions\n\n.. |rtd| image:: https://img.shields.io/readthedocs/skranger.svg\n    :target: http://skranger.readthedocs.io/en/latest/\n\n.. |pypi| image:: https://img.shields.io/pypi/v/skranger.svg\n    :target: https://pypi.python.org/pypi/skranger\n\n.. |pyversions| image:: https://img.shields.io/pypi/pyversions/skranger.svg\n    :target: https://pypi.python.org/pypi/skranger\n\n``skranger`` provides `scikit-learn <https://scikit-learn.org/stable/index.html>`__ compatible Python bindings to the C++ random forest implementation, `ranger <https://github.com/imbs-hl/ranger>`__, using `Cython <https://cython.readthedocs.io/en/latest/>`__.\n\nThe latest release of ``skranger`` uses version `0.12.1 <https://github.com/imbs-hl/ranger/releases/tag/0.12.1>`__ of ``ranger``.\n\n\nInstallation\n------------\n\n``skranger`` is available on `pypi <https://pypi.org/project/skranger>`__ and can be installed via pip:\n\n.. code-block:: bash\n\n    pip install skranger\n\n\nUsage\n-----\n\nThere are two ``sklearn`` compatible classes, ``RangerForestClassifier`` and ``RangerForestRegressor``. There is also the ``RangerForestSurvival`` class, which aims to be compatible with the `scikit-survival <https://github.com/sebp/scikit-survival>`__ API.\n\n\nRangerForestClassifier\n~~~~~~~~~~~~~~~~~~~~~~\n\nThe ``RangerForestClassifier`` predictor uses ``ranger``\'s ForestProbability class to enable both ``predict`` and ``predict_proba`` methods.\n\n.. code-block:: python\n\n    from sklearn.datasets import load_iris\n    from sklearn.model_selection import train_test_split\n    from skranger.ensemble import RangerForestClassifier\n\n    X, y = load_iris(return_X_y=True)\n    X_train, X_test, y_train, y_test = train_test_split(X, y)\n\n    rfc = RangerForestClassifier()\n    rfc.fit(X_train, y_train)\n\n    predictions = rfc.predict(X_test)\n    print(predictions)\n    # [1 2 0 0 0 0 1 2 1 1 2 2 2 1 1 0 1 1 0 1 1 1 0 2 1 0 0 1 2 2 0 1 2 2 0 2 0 0]\n\n    probabilities = rfc.predict_proba(X_test)\n    print(probabilities)\n    # [[0.01333333 0.98666667 0.        ]\n    #  [0.         0.         1.        ]\n    #  ...\n    #  [0.98746032 0.01253968 0.        ]\n    #  [0.99       0.01       0.        ]]\n\n\nRangerForestRegressor\n~~~~~~~~~~~~~~~~~~~~~\n\nThe ``RangerForestRegressor`` predictor uses ``ranger``\'s ForestRegression class. It also supports quantile regression using the ``predict_quantiles`` method.\n\n.. code-block:: python\n\n    from sklearn.datasets import load_boston\n    from sklearn.model_selection import train_test_split\n    from skranger.ensemble import RangerForestRegressor\n\n    X, y = load_boston(return_X_y=True)\n    X_train, X_test, y_train, y_test = train_test_split(X, y)\n\n    rfr = RangerForestRegressor()\n    rfr.fit(X_train, y_train)\n\n    predictions = rfr.predict(X_test)\n    print(predictions)\n    # [26.27401667  8.96549989 24.82981667 27.92506667 28.04606667 45.4693\n    #  21.89681787 40.30345    11.53959613 19.13675    15.88567273 16.69713567\n    #  ...\n    #  20.29025364 26.21245833 23.79643333 14.03546362 21.24893333 34.8825\n    #  21.22463333]\n\n    # enable quantile regression on instantiation\n    rfr = RangerForestRegressor(quantiles=True)\n    rfr.fit(X_train, y_train)\n\n    quantile_lower = rfr.predict_quantiles(X_test, quantiles=[0.1])\n    print(quantile_lower)\n    # [22.    5.   21.88 23.08 23.1  35.89 10.85 31.5   7.04 14.5  11.7  10.9\n    #   8.1  28.38  7.2  19.6  29.1  13.1  24.94 21.09 15.6  11.7  10.41 14.5\n    #  ...\n    #  18.9  21.4   9.43  8.7  26.46 18.99  7.2  19.27 18.5  21.19 18.99 18.88\n    #  14.07 21.87 22.18  9.43 17.28 29.6  18.2 ]\n    quantile_upper = rfr.predict_quantiles(X_test, quantiles=[0.9])\n    print(quantile_upper)\n    # [30.83 12.85 29.01 33.1  33.1  50.   29.75 50.   15.   23.   19.96 21.4\n    #  20.53 50.   13.35 25.   48.5  19.6  46.   26.6  23.7  20.1  17.8  21.4\n    #  ...\n    #  26.78 28.1  17.86 27.5  46.25 24.4  16.74 24.4  28.7  29.1  24.4  25.\n    #  25.   31.51 28.   20.8  26.7  42.13 24.24]\n\n\nRangerForestSurvival\n~~~~~~~~~~~~~~~~~~~~\n\nThe ``RangerForestSurvival`` predictor uses ``ranger``\'s ForestSurvival class, and has an interface similar to the RandomSurvivalForest found in the ``scikit-survival`` package.\n\n.. code-block:: python\n\n    from sksurv.datasets import load_veterans_lung_cancer\n    from sklearn.model_selection import train_test_split\n    from skranger.ensemble import RangerForestSurvival\n\n    X, y = load_veterans_lung_cancer()\n    # select the numeric columns as features\n    X = X[["Age_in_years", "Karnofsky_score", "Months_from_Diagnosis"]]\n    X_train, X_test, y_train, y_test = train_test_split(X, y)\n\n    rfs = RangerForestSurvival()\n    rfs.fit(X_train, y_train)\n\n    predictions = rfs.predict(X_test)\n    print(predictions)\n    # [107.99634921  47.41235714  88.39933333  91.23566667  61.82104762\n    #   61.15052381  90.29888492  47.88706349  21.25111508  85.5768254\n    #   ...\n    #   56.85498016  53.98227381  48.88464683  95.58649206  48.9142619\n    #   57.68516667  71.96549206 101.79123016  58.95402381  98.36299206]\n\n    chf = rfs.predict_cumulative_hazard_function(X_test)\n    print(chf)\n    # [[0.04233333 0.0605     0.24305556 ... 1.6216627  1.6216627  1.6216627 ]\n    #  [0.00583333 0.00583333 0.00583333 ... 1.55410714 1.56410714 1.58410714]\n    #  ...\n    #  [0.12933333 0.14766667 0.14766667 ... 1.64342857 1.64342857 1.65342857]\n    #  [0.00983333 0.0112619  0.04815079 ... 1.79304365 1.79304365 1.79304365]]\n\n    survival = rfs.predict_survival_function(X_test)\n    print(survival)\n    # [[0.95855021 0.94129377 0.78422794 ... 0.19756993 0.19756993 0.19756993]\n    #  [0.99418365 0.99418365 0.99418365 ... 0.21137803 0.20927478 0.20513086]\n    #  ...\n    #  [0.87868102 0.86271864 0.86271864 ... 0.19331611 0.19331611 0.19139258]\n    #  [0.99021486 0.98880127 0.95299007 ... 0.16645277 0.16645277 0.16645277]]\n\n\nLicense\n-------\n\n``skranger`` is licensed under `GPLv3 <https://github.com/crflynn/skranger/blob/master/LICENSE.txt>`__.\n\nDevelopment\n-----------\n\nTo develop locally, it is recommended to have ``asdf``, ``make`` and a C++ compiler already installed. After cloning, run ``make setup``. This will setup the ranger submodule, install python and poetry from ``.tool-versions``, install dependencies using poetry, copy the ranger source code into skranger, and then build and install skranger in the local virtualenv.\n\nTo format code, run ``make fmt``. This will run isort and black against the .py files.\n\nTo run tests and inspect coverage, run ``make test``.\n\nTo rebuild in place after making changes, run ``make build``.\n\nTo create python package artifacts, run ``make dist``.\n\nTo build and view documentation, run ``make docs``.\n',
    'author': 'Flynn',
    'author_email': 'crf204@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/crflynn/skranger',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_is_fitted

from skranger.base import RangerMixin
from skranger.tree._tree import Tree


class BaseRangerTree(RangerMixin, BaseEstimator):
    @property
    def tree_(self):
        check_is_fitted(self)
        return Tree(ranger_forest=self.ranger_forest_)

    def get_depth(self):
        """Calculate the maximum depth of the tree."""
        return self.tree_.get_depth()

    def get_n_leaves(self):
        """Calculate the number of leaves of the tree."""
        return self.tree_.get_n_leaves()

    def apply(self, X):
        """Calculate the index of the leaf for each sample.
        :param array2d X: training input features
        """
        return self.tree_.apply(X)

    def decision_path(self, X):
        """Calculate the decision path through the tree for each sample.
        :param array2d X: training input features
        """
        return self.tree_.decision_path(X)



import warnings
from typing import Iterable

import numpy as np
from sklearn.utils.validation import _check_sample_weight

from skranger import ranger


class RangerMixin:
    @property
    def criterion(self):
        """Compatibility alias for split rule."""
        return self.split_rule

    # region validation
    def _validate_parameters(self, X, y, sample_weights):
        """Validate ranger parameters and set defaults."""
        if hasattr(self, "n_jobs"):
            self.n_jobs_ = max(
                [self.n_jobs, 0]
            )  # sklearn convention is -1 for all, ranger is 0
        self._set_respect_categorical_features()
        self._evaluate_mtry(X.shape[1])
        self._set_importance_mode()
        self.sample_fraction_ = self.sample_fraction or [1.0 if self.replace else 0.632]
        if not isinstance(self.sample_fraction_, Iterable):
            self.sample_fraction_ = [self.sample_fraction_]
        self._check_inbag(sample_weights)
        self._check_set_regularization(X.shape[1])
        self._set_split_rule(y)

    def _evaluate_mtry(self, num_features):
        """Evaluate mtry if callable."""
        if callable(self.mtry):
            self.mtry_ = self.mtry(num_features)
            if self.mtry_ < 1 or self.mtry_ > num_features:
                raise ValueError(
                    "mtry function must evaluate to between 1 and number of features"
                )
        else:
            self.mtry_ = self.mtry
            if self.mtry_ < 0 or self.mtry_ > num_features:
                raise ValueError("mtry must be between 0 and number of features")

    def _set_split_rule(self, y):
        """Set split rule to enum value."""
        if hasattr(self, "save_memory"):
            if (
                self.split_rule == "extratrees"
                and self.respect_categorical_features_ == "partition"
                and self.save_memory
            ):
                raise ValueError(
                    "save memory is not possible with extratrees split rule and unordered predictors"
                )

        if self.num_random_splits > 1 and self.split_rule != "extratrees":
            raise ValueError(
                "random splits must be 1 when split rule is not extratrees"
            )

        if self.tree_type_ in (1, 9):  # classification/probability
            if self.split_rule == "gini":
                self.split_rule_ = 1  # ranger_.SplitRule.LOGRANK
            elif self.split_rule == "extratrees":
                self.split_rule_ = 5  # ranger_.SplitRule.EXTRATREES
            elif self.split_rule == "hellinger":
                if len(np.unique(y)) > 2:
                    raise ValueError(
                        "hellinger split rule can only be used in binary classification"
                    )
                self.split_rule_ = 7  # ranger_.SplitRule.HELLINGER
            else:
                raise ValueError(
                    "split rule must be either gini, extratrees, or hellinger"
                )

        elif self.tree_type_ == 3:  # regression
            if self.split_rule == "variance":
                self.split_rule_ = 1  # ranger_.SplitRule.LOGRANK
            elif self.split_rule == "extratrees":
                self.split_rule_ = 5  # ranger_.SplitRule.EXTRATREES
            elif self.split_rule == "maxstat":
                self.split_rule_ = 4  # ranger_.SplitRule.MAXSTAT
            elif self.split_rule == "beta":
                self.split_rule_ = 6  # ranger_.SplitRule.BETA
                if np.max(y) > 1 or np.max(y) < 0:
                    raise ValueError(
                        "Targets must be between 0 and 1 for beta splitrule"
                    )
            else:
                raise ValueError(
                    "split rule must be either variance, extratrees, maxstat or beta"
                )

        elif self.tree_type_ == 5:  # survival
            if self.split_rule == "logrank":
                self.split_rule_ = 1  # ranger_.SplitRule.LOGRANK
            elif self.split_rule == "extratrees":
                self.split_rule_ = 5  # ranger_.SplitRule.EXTRATREES
            elif self.split_rule == "C":
                self.split_rule_ = 2  # ranger_.SplitRule.AUC
            elif self.split_rule == "C_ignore_ties":
                self.split_rule_ = 3  # ranger_.SplitRule.AUC_IGNORE_TIES
            elif self.split_rule == "maxstat":
                self.split_rule_ = 4  # ranger_.SplitRule.MAXSTAT
            else:
                raise ValueError(
                    "split rule must be either logrank, extratrees, C or maxstat"
                )

    def _check_set_regularization(self, num_features):
        """Check, set the regularization factor to either [] or length num_features."""
        if self.regularization_factor is None:
            self.regularization_factor_ = []
            self.use_regularization_factor_ = False
            return

        if len(self.regularization_factor) > 0:
            if max(self.regularization_factor) > 1:
                raise ValueError("The regularization coefficients must be <= 1")
            if max(self.regularization_factor) <= 0:
                raise ValueError("The regularization coefficients must be > 0")
            if (
                len(self.regularization_factor) != 1
                and len(self.regularization_factor) != num_features
            ):
                raise ValueError(
                    "There must be either one 1 or (number of features) regularization coefficients"
                )
            if len(self.regularization_factor) == 1:
                self.regularization_factor_ = [
                    self.regularization_factor
                ] * num_features

        if all([r == 1 for r in self.regularization_factor]):
            self.regularization_factor_ = []
            self.use_regularization_factor_ = False
        else:
            if getattr(self, "n_jobs_", 1) != 1:
                warnings.warn("Parallelization cannot be used with regularization.")
                self.n_jobs_ = 1
            self.regularization_factor_ = self.regularization_factor
            self.use_regularization_factor_ = True

    def _set_importance_mode(self):
        """Set the importance mode based on ``importance`` and ``local_importance``."""
        # Note IMP_PERM_LIAW is unused
        if self.importance is None or self.importance == "none":
            self.importance_mode_ = 0  # ranger_.ImportanceMode.IMP_NONE
        elif self.importance == "impurity":
            self.importance_mode_ = 1  # ranger_.ImportanceMode.IMP_GINI
        elif (
            self.importance == "impurity_corrected"
            or self.importance == "impurity_unbiased"
        ):
            self.importance_mode_ = 5  # ranger_.ImportanceMode.IMP_GINI_CORRECTED
        elif self.importance == "permutation":
            if self.local_importance:
                self.importance_mode_ = 6  # ranger_.ImportanceMode.IMP_PERM_CASEWISE
            elif self.scale_permutation_importance:
                self.importance_mode_ = 2  # ranger_.ImportanceMode.IMP_PERM_BREIMAN
            else:
                self.importance_mode_ = 3  # ranger_.ImportanceMode.IMP_PERM_RAW
        else:
            raise ValueError("unkown importance mode")

    def _check_inbag(self, sample_weights):
        """Validate related input against ``inbag`` counts."""
        if self.inbag:
            if sample_weights is not None:
                raise ValueError("Cannot use inbag and sample_weights.")
            if len(self.sample_fraction_) > 1:
                raise ValueError("Cannot use class sampling and inbag.")
            if len(self.inbag) != getattr(self, "n_estimators", 1):
                raise ValueError("Size of inbag must be equal to n_estimators.")

    def _check_sample_weight(self, sample_weight, X):
        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, X)
            use_sample_weight = True
            # ranger does additional rng on samples if weights are passed.
            # if the weights are ones, then we dont want that extra rng.
            if np.array_equal(np.unique(sample_weight), np.array([1.0])):
                sample_weight = []
                use_sample_weight = False
        else:
            sample_weight = []
            use_sample_weight = False
        return sample_weight, use_sample_weight

    def _check_split_select_weights(self, split_select_weights):
        if split_select_weights is not None and len(split_select_weights) > 0:
            split_select_weights = np.atleast_2d(split_select_weights).tolist()
            use_split_select_weights = True
        else:
            split_select_weights = [[]]
            use_split_select_weights = False
        return split_select_weights, use_split_select_weights

    def _check_always_split_features(self, always_split_features):
        if always_split_features is not None and len(always_split_features) > 0:
            always_split_features = [str(c).encode() for c in always_split_features]
            use_always_split_features = True
        else:
            always_split_features = []
            use_always_split_features = False
        return always_split_features, use_always_split_features

    def _set_respect_categorical_features(self):
        """Set ``respect_categorical_features`` based on ``split_rule``."""
        if self.respect_categorical_features is None:
            if self.split_rule == "extratrees":
                self.respect_categorical_features_ = "partition"
            else:
                self.respect_categorical_features_ = "ignore"
        else:
            self.respect_categorical_features_ = self.respect_categorical_features

    def _check_categorical_features(self, categorical_features):
        """Determine categorical feature names."""
        use_categorical_features = False
        if self.respect_categorical_features_ == "partition":
            if categorical_features is not None and len(categorical_features) > 0:
                categorical_features = [str(c).encode() for c in categorical_features]
                use_categorical_features = True
        elif (
            self.respect_categorical_features_ == "ignore"
            or self.respect_categorical_features_ == "order"
        ):
            categorical_features = []
            use_categorical_features = False
        else:
            raise ValueError(
                "respect ordered factors must be one of `partition`, `ignore` or `order`"
            )
        if categorical_features is None:
            categorical_features = []
        return categorical_features, use_categorical_features

    # endregion

    # region trees
    def _set_sample_weights(self, sample_weights):
        """Set leaf weights for access in ``Tree``."""
        weights = []
        for tree in sample_weights:
            tree_weights = []
            for idx, node in enumerate(tree):
                tree_weights.append(sum(node))
            weights.append(tree_weights)
        self.ranger_forest_["forest"]["leaf_weights"] = weights

    def _get_sample_values(self, values):
        """Map the leaf samples to corresponding values.

        Create a similarly structured set of lists by mapping the leaf sample indexes
        to corresponding values passed. Used for mapping leaf samples to target values
        or sample weights.
        """
        mapped_values = []
        for tree in self.ranger_forest_["forest"]["leaf_samples"]:
            mapped_values.append([])
            for node in tree:
                mapped_values[-1].append([values[idx] for idx in node])
        return mapped_values

    def _get_values(self, left, right, idx, values):
        """Recursively sum the child values through a tree."""
        left_values = (
            values[idx]
            if left[idx] == 0
            else self._get_values(left, right, left[idx], values)
        )
        right_values = (
            values[idx]
            if right[idx] == 0
            else self._get_values(left, right, right[idx], values)
        )
        values[idx] = left_values + right_values
        return values[idx]

    def _set_node_values(self, y, sample_weight):
        """Set the node values for ``Tree.value``."""
        forest_values = self._get_sample_values(y)
        forest_weights = self._get_sample_values(sample_weight)
        self._set_sample_weights(forest_weights)
        self.ranger_forest_["forest"]["node_values"] = []
        for idx in range(self.ranger_forest_["num_trees"]):
            left = self.ranger_forest_["forest"]["child_node_ids"][idx][0]
            right = self.ranger_forest_["forest"]["child_node_ids"][idx][1]
            root = 0
            values = forest_values[idx]
            self._get_values(
                left, right, root, values,
            )
            weights = forest_weights[idx]
            self._get_values(
                left, right, root, weights,
            )
            if getattr(self, "_estimator_type", None) == "classifier":
                values = [
                    np.bincount(v, weights=w, minlength=self.n_classes_).tolist()
                    if v
                    else [np.nan] * self.n_classes_
                    for v, w in zip(values, weights)
                ]
            else:
                values = [
                    np.average(v, weights=w, axis=0) if v else np.nan
                    for v, w in zip(values, weights)
                ]
            self.ranger_forest_["forest"]["node_values"].append(values)

    def _set_n_classes(self):
        """Set num classes for ``Tree.n_classes``."""
        # for accessing in Tree
        self.ranger_forest_["n_classes"] = getattr(self, "n_classes_", 1)

    def _set_leaf_samples(self, terminal_nodes):
        """Set the leaf samples using the terminal nodes.

        Collect all of the record indexes that fall into each terminal node. The
        resulting ``leaf_samples`` value will be a list of lists of lists. The outer
        list is a collection of tree lists. The tree lists are collections of node
        lists. The node lists contain record indexes for terminal nodes only. The index
        of the node lists correspond to the node index in the respective tree.
        """
        leaf_samples = []
        for tree_idx, tree in enumerate(terminal_nodes.T):
            n_nodes = len(self.ranger_forest_["forest"]["child_node_ids"][tree_idx][0])
            tree_leaf_samples = [[] for _ in range(n_nodes)]
            for record_idx, terminal_node in enumerate(tree):
                tree_leaf_samples[terminal_node].append(record_idx)
            leaf_samples.append(tree_leaf_samples)
        self.ranger_forest_["forest"]["leaf_samples"] = leaf_samples

    def _get_terminal_node_forest(self, X):
        """Get a terminal node forest for X.

        Uses a trained forest to return the terminal node ids of each record of ``X``
        for each tree. Returns a dictionary.

        The returned value of key ``predictions`` will hold a list of
        lists. The inner list is the list of terminal nodes of each tree for a record.
        The outer list entries correspond to each record of ``X``.

        :param array2d X: prediction input features
        """
        # many fields defaulted here which are unused
        forest = ranger.ranger(
            self.tree_type_,
            np.asfortranarray(X.astype("float64")),
            np.asfortranarray([[]]),
            self.feature_names_,  # variable_names
            0,  # m_try
            getattr(self, "n_estimators", 1),  # num_trees
            self.verbose,
            self.seed,
            getattr(self, "n_jobs_", 1),  # num_threads
            False,  # write_forest
            0,  # importance_mode
            0,  # min_node_size
            [],  # split_select_weights
            False,  # use_split_select_weights
            [],  # always_split_feature_names
            False,  # use_always_split_feature_names
            True,  # prediction_mode
            self.ranger_forest_["forest"],  # loaded_forest
            True,  # sample_with_replacement
            False,  # probability
            [],  # unordered_feature_names
            False,  # use_unordered_features
            False,  # save_memory
            1,  # split_rule
            [],  # case_weights
            False,  # use_case_weights
            {},  # class_weights
            False,  # predict_all
            self.keep_inbag,
            [1],  # sample_fraction
            0,  # alpha
            0,  # minprop
            self.holdout,
            2,  # prediction_type (terminal nodes)
            1,  # num_random_splits
            False,  # oob_error
            0,  # max_depth
            [],  # inbag
            False,  # use_inbag
            [],  # regularization_factor_
            False,  # use_regularization_factor_
            False,  # regularization_usedepth
        )
        return forest

    # endregion
